import re
import shutil
import sqlite3
import pandas as pd

from pathlib import Path


MLOG_DIR = Path("./mlog")
MLOG_DB = MLOG_DIR / "mlog.db"
KEY_FORMAT = "[a-zA-Z][a-zA-Z0-9_]*"
GET_FORMAT = "[a-zA-Z_][a-zA-Z0-9_]*"

SQL_CREATE_RUNS_TABLE = """
CREATE TABLE IF NOT EXISTS runs (
    _id INTEGER PRIMARY KEY AUTOINCREMENT,
    _name VARCHAR(255),
    _source VARCHAR(255)
)
"""

SQL_CREATE_LOGS_TABLE = """
CREATE TABLE IF NOT EXISTS logs (
    _id INTEGER PRIMARY KEY AUTOINCREMENT,
    _run_id INT,
    FOREIGN KEY (_run_id) REFERENCES runs (_id)
)
"""

def start(run=None, config=None, save=None):
    return Run(run=run, config=config, save=save)


def get(*columns, filters=None):

    con = sqlite3.connect(MLOG_DB)

    for column in columns:
        if not re.fullmatch(GET_FORMAT, column):
            raise ValueError(
                f"Column '{column}' does not use format '{GET_FORMAT}'")

    columns = list(columns)
    columns.append('_id')
    columns = ",".join(columns)

    data = pd.read_sql_query(f"SELECT {columns} FROM logs", con)

    con.close()

    return data.set_index('_id')


class Run:

    def __init__(self, run=None, config=None, save=None):

        MLOG_DIR.mkdir(parents=True, exist_ok=True)

        if run is not None and not re.fullmatch(KEY_FORMAT, run):
            raise ValueError(
                f"Run name '{run}' does not use format '{KEY_FORMAT}'")

        con = sqlite3.connect(MLOG_DB)
        cur = con.cursor()

        if config:

            # Retrieve existing columns
            cols = [col[1] for col in cur.execute('PRAGMA table_info(runs)')]

            # Check columns format and add missing columns
            for key in config.keys():

                if not re.fullmatch(KEY_FORMAT, key):
                    raise ValueError(
                        f"Column '{key}' does not use format '{KEY_FORMAT}'")

                if key not in cols:
                    cur.execute(f"ALTER TABLE runs ADD {key}")

            # Add name
            config["_name"] = run

            # Add configs
            cols = ",".join(config.keys())
            vals = ":" + ",:".join(config.keys())
            cur.execute(f"INSERT INTO runs ({cols}) VALUES ({vals})", config)

            # Remove name
            config.pop("_name")

        else:
            cur.execute("INSERT INTO runs DEFAULT VALUES")

        self.run_id = cur.lastrowid

        con.commit()
        con.close()

        # Save files
        if save is not None:

            save_directory = MLOG_DIR / str(self.run_id)
            save_directory.mkdir()

            for file in Path('.').glob(save):
                shutil.copy(file, save_directory)

    def log(self, **logs):

        con = sqlite3.connect(MLOG_DB)

        with con:
            # Retrieve existing columns
            cols = [col[1] for col in con.execute("PRAGMA table_info(logs)")]

            # Check columns and values format and add missing columns
            for key, val in logs.items():

                if not re.fullmatch(KEY_FORMAT, key):
                    raise ValueError(
                        f"Column '{key}' does not use format '{KEY_FORMAT}'")

                if key not in cols:
                    con.execute(f"ALTER TABLE logs ADD {key} REAL")

                try:
                    float(val)
                except ValueError:
                    raise ValueError(
                        f"Value '{val}' for column '{key}' is not a number")

            # Add run id
            logs['_run_id'] = self.run_id

            # Add logs
            cols = ",".join(logs.keys())
            vals = ":" + ",:".join(logs.keys())
            con.execute(f"INSERT INTO logs ({cols}) VALUES ({vals})", logs)

        # Remove run id
        logs.pop('_run_id')

        con.commit()
        con.close()

    def get(self, *columns):
        # TODO: call get with appropriate filter

        con = sqlite3.connect(MLOG_DB)

        for column in columns:
            if not re.fullmatch(GET_FORMAT, column):
                raise ValueError(
                    f"Column '{column}' does not use format '{GET_FORMAT}'")

        columns = list(columns)
        columns.append('_id')
        columns = ",".join(columns)

        data = pd.read_sql_query(
            f"SELECT {columns} FROM logs WHERE _run_id = '{self.run_id}'", con)

        con.close()

        return data.set_index('_id')



from pathlib import Path

import matplotlib.pyplot as plt
import mlog

from argparse import ArgumentParser


AGGS = ['mean', 'median']
INTS = ['std', 'max']

REDUCTIONS = ['mean', 'median', 'min', 'max', 'std']


def plot(args):

    df = mlog.get('_run_id', args.x_axis, args.y_axis)

    fig, ax = plt.subplots()

    if args.scatter:
        ax.scatter(df[args.x_axis], df[args.y_axis])

    elif args.ungroup:
        for run_id, run in df.groupby('_run_id'):
            ax.plot(run[args.x_axis], run[args.y_axis])

    else:
        df = df.groupby(args.x_axis).agg(REDUCTIONS)

        if args.intervals == 'max':
            df['_lower'] = df[args.y_axis]['min']
            df['_upper'] = df[args.y_axis]['max']
        elif args.intervals == 'std':
            df['_lower'] = df[args.y_axis]['mean'] - df[args.y_axis]['std']
            df['_upper'] = df[args.y_axis]['mean'] + df[args.y_axis]['std']

        ax.plot(df.index, df[args.y_axis][args.aggregate])
        ax.fill_between(df.index, df['_lower'], df['_upper'], alpha=0.4)

    ax.set_xlabel(args.x_axis)
    ax.set_ylabel(args.y_axis)

    plt.tight_layout()

    if args.output is not None:
        plt.savefig(args.output)

    plt.show()


def main():

    parser = ArgumentParser()
    subparsers = parser.add_subparsers(dest='command', required=True)

    # Plot
    parser_plot = subparsers.add_parser('plot')
    parser_plot.set_defaults(func=plot)
    parser_plot.add_argument('x_axis')
    parser_plot.add_argument('y_axis')

    parser_plot.add_argument('-s', '--scatter', action='store_true')
    parser_plot.add_argument('-u', '--ungroup', action='store_true')
    parser_plot.add_argument('-a', '--aggregate', choices=AGGS, default='mean')
    parser_plot.add_argument('-i', '--intervals', choices=INTS, default='std')

    parser_plot.add_argument('-o', '--output')

    # Parse arguments and execute command
    args = parser.parse_args()
    args.func(args)



import sqlite3

from .mlog import start, get
from .mlog import MLOG_DB, SQL_CREATE_RUNS_TABLE, SQL_CREATE_LOGS_TABLE


con = sqlite3.connect(MLOG_DB)

with con:
    con.execute(SQL_CREATE_RUNS_TABLE)
    con.execute(SQL_CREATE_LOGS_TABLE)

con.close()
